from ima_predeployed.addresses import MESSAGE_PROXY_FOR_SCHAIN_ADDRESS, TOKEN_MANAGER_ERC20_ADDRESS, \
    TOKEN_MANAGER_ERC721_ADDRESS, TOKEN_MANAGER_ETH_ADDRESS, TOKEN_MANAGER_ERC1155_ADDRESS
from web3 import Web3
from ..contract_generator import ContractGenerator, next_slot


class TokenManagerLinkerGenerator(ContractGenerator):
    ARTIFACT_FILENAME = "TokenManagerLinker.json"
    DEFAULT_ADMIN_ROLE = (0).to_bytes(32, 'big')
    REGISTRAR_ROLE = Web3.solidityKeccak(['string'], ['REGISTRAR_ROLE'])

    # ---------- storage ----------
    # --------Initializable--------
    # 0:    _initialized, _initializing;
    # -----ContextUpgradeable------
    # 1:    __gap
    # ...   __gap
    # 50:   __gap
    # ------ERC165Upgradeable------
    # 51:   __gap
    # ...   __gap
    # 100:  __gap
    # --AccessControlUpgradeable---
    # 101:  _roles
    # 102:  __gap
    # ...   __gap
    # 150:  __gap
    # AccessControlEnumerableUpgradeable
    # 151:  _roleMembers
    # 152:  __gap
    # ...   __gap
    # 200:  __gap
    # ------TokenManagerLinker------
    # 201:  messageProxy
    # 202:  linkerAddress
    # 203:  tokenManagers
    # 204:  interchainConnections

    INITIALIZED_SLOT = 0
    ROLES_SLOT = 101
    ROLE_MEMBERS_SLOT = 151
    MESSAGE_PROXY_SLOT = 201
    LINKER_ADDRESS_SLOT = next_slot(MESSAGE_PROXY_SLOT)
    TOKEN_MANAGERS_SLOT = next_slot(LINKER_ADDRESS_SLOT)

    def __init__(self, deployer_address: str, linker_address: str):
        super().__init__(self.ARTIFACT_FILENAME)
        self._setup(deployer_address, linker_address)

    # private

    def _setup(self, deployer_address: str, linker_addres: str) -> None:
        self._write_uint256(self.INITIALIZED_SLOT, 1)
        self._setup_role(self.ROLES_SLOT, self.ROLE_MEMBERS_SLOT, self.DEFAULT_ADMIN_ROLE, [deployer_address])
        self._setup_role(self.ROLES_SLOT, self.ROLE_MEMBERS_SLOT, self.REGISTRAR_ROLE, [deployer_address])
        self._write_address(self.MESSAGE_PROXY_SLOT, MESSAGE_PROXY_FOR_SCHAIN_ADDRESS)
        self._write_address(self.LINKER_ADDRESS_SLOT, linker_addres)
        self._write_addresses_array(
            self.TOKEN_MANAGERS_SLOT, [
                TOKEN_MANAGER_ETH_ADDRESS,
                TOKEN_MANAGER_ERC20_ADDRESS,
                TOKEN_MANAGER_ERC721_ADDRESS,
                TOKEN_MANAGER_ERC1155_ADDRESS])



from ..contract_generator import ContractGenerator, calculate_mapping_value_slot, calculate_array_value_slot, next_slot
from ..addresses import COMMUNITY_LOCKER_ADDRESS, KEY_STORAGE_ADDRESS, TOKEN_MANAGER_ERC1155_ADDRESS,\
    TOKEN_MANAGER_ERC20_ADDRESS, TOKEN_MANAGER_ERC721_ADDRESS, TOKEN_MANAGER_ETH_ADDRESS
from web3 import Web3
from pkg_resources import get_distribution

class MessageProxyForSchainGenerator(ContractGenerator):
    ARTIFACT_FILENAME = "MessageProxyForSchain.json"
    DEFAULT_ADMIN_ROLE = (0).to_bytes(32, 'big')
    MAINNET_HASH = Web3.solidityKeccak(['string'], ['Mainnet'])
    GAS_LIMIT = 3000000
    ANY_SCHAIN = (0).to_bytes(32, 'big')

    # ---------- storage ----------
    # --------Initializable--------
    # 0:    _initialized, _initializing;
    # -----ContextUpgradeable------
    # 1:    __gap
    # ...   __gap
    # 50:   __gap
    # ------ERC165Upgradeable------
    # 51:   __gap
    # ...   __gap
    # 100:  __gap
    # --AccessControlUpgradeable---
    # 101:  _roles
    # 102:  __gap
    # ...   __gap
    # 150:  __gap
    # AccessControlEnumerableUpgradeable
    # 151:  _roleMembers
    # 152:  __gap
    # ...   __gap
    # 200:  __gap
    # ---------MessageProxy--------
    # 201:  connectedChains
    # 202:  deprecatedRegistryContracts
    # 203:  gasLimit
    # ----MessageProxyForSchain----
    # 204:  keyStorage
    # 205:  schainHash
    # 206:  _outgoingMessageDataHash
    # 207:  _idxHead
    # 208:  _idxTail
    # 209:  _registryContracts
    # 210:  version

    INITIALIZED_SLOT = 0
    ROLES_SLOT = 101
    ROLE_MEMBERS_SLOT = 151
    CONNECTED_CHAINS_SLOT = 201
    DEPRECATED_REGISTRY_CONTRACTS_SLOT = next_slot(CONNECTED_CHAINS_SLOT)
    GAS_LIMIT_SLOT = next_slot(DEPRECATED_REGISTRY_CONTRACTS_SLOT)
    KEY_STORAGE_SLOT = next_slot(GAS_LIMIT_SLOT)
    SCHAIN_HASH_SLOT = next_slot(KEY_STORAGE_SLOT)
    OUTGOING_MESSAGE_DATA_HASH = next_slot(SCHAIN_HASH_SLOT)
    IDX_HEAD = next_slot(OUTGOING_MESSAGE_DATA_HASH)
    IDX_TAIL = next_slot(IDX_HEAD)
    REGISTRY_CONTRACTS_SLOT = next_slot(IDX_TAIL)
    VERSION_SLOT = next_slot(REGISTRY_CONTRACTS_SLOT)
    

    def __init__(self, deployer_address: str, schain_name: str):
        super().__init__(self.ARTIFACT_FILENAME)
        self._setup(deployer_address, schain_name)

    # private

    def _setup(self, deployer_address: str, schain_name: str) -> None:
        self._write_uint256(self.INITIALIZED_SLOT, 1)
        self._setup_role(self.ROLES_SLOT, self.ROLE_MEMBERS_SLOT, self.DEFAULT_ADMIN_ROLE, [deployer_address])
        self._write_address(self.KEY_STORAGE_SLOT, KEY_STORAGE_ADDRESS)
        self._write_bytes32(self.SCHAIN_HASH_SLOT, Web3.solidityKeccak(['string'], [schain_name]))

        connected_chain_info_slot = calculate_mapping_value_slot(
            self.CONNECTED_CHAINS_SLOT, self.MAINNET_HASH, 'bytes32')
        inited_slot = connected_chain_info_slot + 2
        self._write_uint256(inited_slot, 1)
        self._write_uint256(self.GAS_LIMIT_SLOT, self.GAS_LIMIT)
        self._write_string(self.VERSION_SLOT, get_distribution('ima_predeployed').version)
        registry_contracts_slot = calculate_mapping_value_slot(
            self.REGISTRY_CONTRACTS_SLOT, self.ANY_SCHAIN, 'bytes32')
        allowed_contracts = [
            TOKEN_MANAGER_ETH_ADDRESS,
            TOKEN_MANAGER_ERC20_ADDRESS,
            TOKEN_MANAGER_ERC721_ADDRESS,
            TOKEN_MANAGER_ERC1155_ADDRESS,
            COMMUNITY_LOCKER_ADDRESS]
        values_slot = registry_contracts_slot
        indexes_slot = registry_contracts_slot + 1
        self._write_uint256(values_slot, len(allowed_contracts))
        for i, contract in enumerate(allowed_contracts):
            self._write_address(calculate_array_value_slot(values_slot, i), contract)
            self._write_uint256(calculate_mapping_value_slot(indexes_slot, int(contract, 16), 'uint256'), i + 1)



from ima_predeployed.addresses import TOKEN_MANAGER_ETH_ADDRESS
from ima_predeployed.contract_generator import ContractGenerator, next_slot
from web3 import Web3


class EthErc20Generator(ContractGenerator):
    ARTIFACT_FILENAME = "EthErc20.json"
    DEFAULT_ADMIN_ROLE = (0).to_bytes(32, 'big')
    MINTER_ROLE = Web3.solidityKeccak(['string'], ['MINTER_ROLE'])
    BURNER_ROLE = Web3.solidityKeccak(['string'], ['BURNER_ROLE'])
    NAME = 'ERC20 Ether Clone'
    SYMBOL = 'ETHC'
    DECIMALS = 18

    # ---------- storage ----------
    # --------Initializable--------
    # 0:    _initialized, _initializing;
    # -----ContextUpgradeable------
    # 1:    __gap
    # ...   __gap
    # 50:   __gap
    # ------ERC165Upgradeable------
    # 51:   __gap
    # ...   __gap
    # 100:  __gap
    # --AccessControlUpgradeable---
    # 101:  _roles
    # 102:  __gap
    # ...   __gap
    # 150:  __gap
    # AccessControlEnumerableUpgradeable
    # 151:  _roleMembers
    # 152:  __gap
    # ...   __gap
    # 200:  __gap
    # -------ERC20Upgradeable-------
    # 201:  _balances
    # 202:  _allowances
    # 203:  _totalSupply
    # 204:  _name
    # 205:  _symbol
    # 206:  _decimals
    # 207:  __gap
    # ...   __gap
    # 250:  __gap
    # ---ERC20BurnableUpgradeable---
    # 251:  __gap
    # ...   __gap
    # 300:  __gap
    # -----------EthErc20-----------

    INITIALIZED_SLOT = 0
    ROLES_SLOT = 101
    ROLE_MEMBERS_SLOT = 151
    NAME_SLOT = 204
    SYMBOL_SLOT = next_slot(NAME_SLOT)
    DECIMALS_SLOT = next_slot(SYMBOL_SLOT)

    def __init__(self, deployer_address: str):
        super().__init__(self.ARTIFACT_FILENAME)
        self._setup(deployer_address)

    # private

    def _setup(self, deployer_address: str) -> None:
        self._write_uint256(self.INITIALIZED_SLOT, 1)
        self._setup_role(self.ROLES_SLOT, self.ROLE_MEMBERS_SLOT, self.DEFAULT_ADMIN_ROLE, [deployer_address])
        self._setup_role(self.ROLES_SLOT, self.ROLE_MEMBERS_SLOT, self.MINTER_ROLE, [TOKEN_MANAGER_ETH_ADDRESS])
        self._setup_role(self.ROLES_SLOT, self.ROLE_MEMBERS_SLOT, self.BURNER_ROLE, [TOKEN_MANAGER_ETH_ADDRESS])
        self._write_string(self.NAME_SLOT, self.NAME)
        self._write_string(self.SYMBOL_SLOT, self.SYMBOL)
        self._write_uint256(self.DECIMALS_SLOT, self.DECIMALS)



import json
import os
from web3 import Web3


def to_even_length(hex_string: str) -> str:
    assert hex_string.startswith('0x')
    if len(hex_string) % 2 != 0:
        return "0x0" + hex_string[2:]
    else:
        return hex_string


def add_0x(bytes_string: str) -> str:
    if bytes_string.startswith('0x'):
        return bytes_string
    else:
        return '0x' + bytes_string


def calculate_mapping_value_slot(slot: int, key: any, key_type: str) -> int:
    if key_type == 'address':
        return calculate_mapping_value_slot(slot, int(key, 16).to_bytes(32, 'big'), 'bytes32')
    else:
        return int.from_bytes(Web3.solidityKeccak([key_type, 'uint256'], [key, slot]), 'big')


def calculate_array_value_slot(slot: int, index: int) -> int:
    return int.from_bytes(Web3.solidityKeccak(['uint256'], [slot]), 'big') + index


def next_slot(previous_slot: int) -> int:
    return previous_slot + 1


class ContractGenerator:
    def __init__(self, artifact_filename: str):
        artifacts_dir = os.path.join(os.path.dirname(__file__), 'artifacts')
        artifacts_path = os.path.join(artifacts_dir, artifact_filename)
        with open(artifacts_path, encoding='utf-8') as fp:
            contract = json.load(fp)
            self.bytecode = contract['deployedBytecode']
            self.abi = contract['abi']
            self.storage = {}

    def generate_contract(self, balance: int = 0, nonce: int = 0) -> dict:
        assert isinstance(self.bytecode, str)
        assert isinstance(self.storage, dict)
        return {
            'code': self.bytecode,
            'balance': str(balance),
            'nonce': str(nonce),
            'storage': self.storage
        }

    # private

    def _write_address(self, slot: int, address: str) -> None:
        self.storage[to_even_length(hex(slot))] = address.lower()

    def _write_bytes32(self, slot: int, data: bytes) -> None:
        assert len(data) <= 32
        self.storage[to_even_length(hex(slot))] = to_even_length(add_0x(data.hex()))

    def _write_uint256(self, slot: int, value: int) -> None:
        self.storage[to_even_length(hex(slot))] = to_even_length(add_0x(hex(value)))

    def _setup_role(self, roles_slot: int, role_members_slot: int, role: bytes, accounts: [str]):
        role_data_slot = calculate_mapping_value_slot(roles_slot, role, 'bytes32')
        members_slot = role_data_slot
        role_members_value_slot = calculate_mapping_value_slot(role_members_slot, role, 'bytes32')
        values_slot = role_members_value_slot
        indexes_slot = role_members_value_slot + 1
        self._write_uint256(values_slot, len(accounts))
        for i, account in enumerate(accounts):
            members_value_slot = calculate_mapping_value_slot(members_slot, account, 'address')
            self._write_uint256(members_value_slot, 1)
            self._write_address(calculate_array_value_slot(values_slot, i), account)
            self._write_uint256(calculate_mapping_value_slot(indexes_slot, int(account, 16), 'uint256'), i + 1)

    def _write_addresses_array(self, slot: int, values: list) -> None:
        self._write_uint256(slot, len(values))
        for i, address in enumerate(values):
            address_slot = calculate_array_value_slot(slot, i)
            self._write_address(address_slot, address)

    def _write_string(self, slot: int, value: str) -> None:
        binary = value.encode()
        length = len(binary)
        if length < 32:
            binary += (2 * length).to_bytes(32 - length, 'big')
            self._write_bytes32(slot, binary)
        else:
            self._write_uint256(slot, 2 * length + 1)

            def chunks(size, source):
                for i in range(0, len(source), size):
                    yield source[i:i + size]

            for index, data in enumerate(chunks(32, binary)):
                if len(data) < 32:
                    data += int(0).to_bytes(32 - len(data), 'big')
                self._write_bytes32(calculate_array_value_slot(slot, index), data)
